from __future__ import annotations

from collections import defaultdict
import orjson
import polars as pl

from .logic import ExpressionBuilder
from .entities import Node, TargetStats
from .usecases import find_split_cat, find_split_num, read_data, get_col


class DecisionTree:
    """
    Implementation of the CART regression tree with few tweaks for robustness.

    The tweaks include:
    * Considering only balanced splits for numerical variables.
    * Ignoring non-informative categories.
    * Handling missing values in numerical and categorical variables.
    * Handling variables with constant values.

    The convention: right condition is always precise and does not include missing values,
        unless it has missing value explicitly on the RHS.
    """

    def __init__(self, max_depth: int = 4, min_leaf_size: int = 10) -> None:
        self.max_depth = max(max_depth, 2)  # ignore values < 2
        self.min_leaf_size = max(min_leaf_size, 1)  # ignore values < 1
        self.root_ = None
        self.feature_importances_ = None
        self.prediction_type_ = None

    @property
    def leaves(self) -> list[Node]:
        if not self.root_:  # if untrained tree.
            return []

        return [node for node in self.get_nodes() if node.left is None and node.right is None]

    def get_nodes(self, start: Node | None = None) -> list[Node]:
        """
        Get tree nodes starting from a given node. If no starting node is provided,
            then `root_` node is used as starting node.

        Returns
        -------
        list[Node]
            Flat list of tree nodes.
        """

        start = start or self.root_
        if not start:
            return []

        left = self.get_nodes(start.left) if start.left else []
        right = self.get_nodes(start.right) if start.right else []
        return [start] + left + right

    def fit(self, data: pl.LazyFrame | pl.DataFrame | str, y_true: pl.Series | str) -> DecisionTree:
        """
        Args
        ----
        data : pl.LazyFrame | pl.DataFrame
        y_true : pl.Series
        """

        # prepare data and target
        if isinstance(data, str) and isinstance(y_true, str):
            data = read_data(data)
            y_true = get_col(data, y_true)

        if y_true.name in data.columns:
            data = data.drop(y_true.name)

        # validate settings.
        if self.min_leaf_size >= len(y_true):
            raise RuntimeError("min_leaf_size must be smaller than data size")

        # training.
        self.prediction_type_ = pl.Float64 if y_true.dtype.is_numeric() else pl.String
        self.root_ = self._make_node(node=Node(depth=1), data=data, y_true=y_true)

        # postprocessing: compute `feature_importances_`
        self._compute_feature_importance()
        return self

    def _compute_feature_importance(self):
        self.feature_importances_ = defaultdict(lambda: 0)  # 0 -- is a default value
        for node in self.get_nodes():
            if node.condition is None:
                continue
            # when node is split the split stats are written to the *parent's* `target_stats`
            self.feature_importances_[node.condition.column] += node.size * node.target_stats.score_reduction

        total_score = sum(self.feature_importances_.values())
        self.feature_importances_ = {
            colname: score / total_score
            for colname, score
            in sorted(self.feature_importances_.items(), key=lambda item: item[1], reverse=True)
        }

    def _make_node(self, node: Node, data: pl.LazyFrame | pl.DataFrame, y_true: pl.Series) -> Node | None:
        """
        Finalizes tree node, provided on input.

        Invariants:
        * any non-terminal node has defined split condition.
        * any node has information about target statistics.
        * either both `left` and `right` child nodes are defined or both are `None`
        * `right` child corresponds to the samples, where `condition = True` and `left` child to all remaining samples.
        """

        def maybe_get_in_memory_df() -> pl.DataFrame:
            """
            As we go further down the tree, the input data for each node becomes smaller and smaller.
            At some point it becomes small enough to safely use in-memory representation instead of the LazyFrame.
            """

            if isinstance(data, pl.DataFrame):
                return data

            MAX_IN_MEMORY_SIZE = 20_000  # convert to in-memory data frame  if the number of samples is below threshold.
            return data.collect() if len(y_true) <= MAX_IN_MEMORY_SIZE else data

        def init_node_stats():
            node.size = len(y_true)
            node.target_stats = TargetStats()
            if y_true.dtype.is_numeric():
                node.target_stats.mean = y_true.mean()
                node.target_stats.var = y_true.var()
            else:
                node.target_stats.distr = {cat: count for cat, count in y_true.value_counts().rows()}

        init_node_stats()

        if node.depth > self.max_depth:
            return None  # max depth reached.

        if len(y_true) < self.min_leaf_size:
            return None  # insufficient data.

        data = maybe_get_in_memory_df()

        # find the best split
        best_split = None
        for colname in data.columns:
            if data.schema[colname].is_numeric():
                cur_split = find_split_num(data=data, colname=colname, y_true=y_true)
            else:
                cur_split = find_split_cat(data=data, colname=colname, y_true=y_true)
            if cur_split.best_idx is None:
                continue

            if best_split is None or best_split.best_split_eval < cur_split.best_split_eval:
                best_split = cur_split

        if best_split is None or best_split.best_split_eval <= 0:
            return node

        # note: split score is written to the *parent* node!
        node.target_stats.score_reduction = best_split.best_split_eval

        # make conditions leading to the right and left branches
        right_condition = best_split.best_split_condition
        left_condition = ExpressionBuilder(best_split.best_split_condition).not_().current

        # make child nodes
        right_mask = right_condition.apply(data)
        left_mask = left_condition.apply(data)
        node.right = self._make_node(
            node=Node(depth=node.depth+1, parent=node),
            data=data.filter(right_mask),
            y_true=y_true.filter(right_mask)
        )
        node.left = self._make_node(
            node=Node(depth=node.depth+1, parent=node),
            data=data.filter(left_mask),
            y_true=y_true.filter(left_mask)
        )

        # either both left and right nodes are present, or both are absent. Otherwise, the prediction is not possible.
        node.left = node.left if node.right is not None else None
        node.right = node.right if node.left is not None else None

        # node's condition is defined only if it is a non-terminal node.
        if node.right is None and node.left is None:
            node.condition = None
        else:
            node.condition = right_condition

        return node

    def predict(self, data: pl.LazyFrame) -> pl.Series:
        res = None
        for leaf in self.leaves:
            cur_flag = leaf.full_condition.apply(data)

            if res is None:
                res = pl.Series(values=[None] * len(cur_flag), dtype=self.prediction_type_)

            if self.prediction_type_.is_numeric():
                res = res.set(cur_flag, leaf.target_stats.mean)
            else:
                max_class, max_count = max(leaf.target_stats.distr.items(), key=lambda item: item[1])
                res = res.set(cur_flag, max_class)
        return res

    def save(self, filename: str):
        with open(filename, "wb") as f:
            f.write(orjson.dumps(
                {
                    "root_": self.root_.serialize() if self.root_ is not None else None,
                    "max_depth": self.max_depth,
                    "min_leaf_size": self.min_leaf_size,
                    "feature_importances_": self.feature_importances_,
                    "prediction_type_": str(self.prediction_type_) if self.prediction_type_ else None,
                },
                option=orjson.OPT_SERIALIZE_NUMPY)
            )

    @staticmethod
    def load(filename: str) -> DecisionTree:
        with open(filename, "rb") as f:
            data = orjson.loads(f.read())

        res = DecisionTree(
            max_depth=data["max_depth"],
            min_leaf_size=data["min_leaf_size"],
        )
        res.root_ = Node.deserialize(data["root_"]) if data["root_"] is not None else None
        res.feature_importances_ = data["feature_importances_"]

        pred_types = {"Float64": pl.Float64, "String": pl.String, None: None}
        res.prediction_type_ = pred_types[data["prediction_type_"]]
        return res



from __future__ import annotations

import numpy as np
import polars as pl

from .logic import Operator, AtomicExpression, ExpressionBuilder
from .entities import SplitReport, BaseSplitScoring, VarianceScoring, EntropyScoring


def read_data(filename: str) -> pl.LazyFrame:
    if filename.endswith(".csv"):
        return pl.scan_csv(filename)
    if filename.endswith(".parquet") or filename.endswith(".pq") or filename.endswith(".pqt"):
        return pl.scan_parquet(filename)
    raise NotImplementedError(f"Unknown data file format: {filename}")


def get_col(data: pl.LazyFrame | pl.DataFrame, colname: str) -> pl.Series:
    if isinstance(data, pl.DataFrame):
        return data[colname]
    else:
        return data.select(colname).collect().to_series()


def get_scoring_method(data: pl.LazyFrame | pl.DataFrame, y_true: pl.Series, colname: str) -> BaseSplitScoring:
    if y_true.dtype.is_numeric():
        return VarianceScoring(data=data, y_true=y_true, column=colname)
    return EntropyScoring(data=data, y_true=y_true, column=colname)


def find_split_num(
    data: pl.LazyFrame | pl.DataFrame,
    colname: str,
    y_true: pl.Series,
    scoring: BaseSplitScoring | None = None
) -> SplitReport:
    """
    Heuristic method for finding a split point, that maximizes variance reduction after the split.

    The candidate points, which are checked:
    * 25, 50, 75 percentiles
    * np.nan -- if missing values are present
    """

    class ColStats:
        """
        Wrapper for the polars `describe` result.
        """

        def __init__(self, column: pl.Series):
            self.stats = column.describe()

        @property
        def is_dull(self) -> bool:
            is_all_missing = len(self.stats) == 2
            is_all_nan = not is_all_missing and np.isnan(self.std)
            is_constant = not is_all_missing and self.std == 0
            return is_all_missing or is_constant or is_all_nan

        @property
        def has_missing_values(self) -> bool:
            null_count_idx = 1  # index at which the statistics contain information about missing values.
            return self.stats["value"][null_count_idx] > 0

        @property
        def std(self) -> float:
            return self.stats["value"][3]

        def perc(self, perc_name: str) -> float:
            idx = self.stats["statistic"].index_of(perc_name)
            return self.stats["value"][idx]

    scoring = scoring or get_scoring_method(data=data, y_true=y_true, colname=colname)

    col = get_col(data, colname)
    if not col.dtype.is_numeric():
        raise TypeError(f"{colname} is not numeric or all-None.")

    col_stats = ColStats(col)
    if col_stats.is_dull:
        return scoring.get_report()

    for perc_name in ["25%", "50%", "75%"]:
        split_pt = col_stats.perc(perc_name)
        split_condition = ExpressionBuilder(
            AtomicExpression(colname=colname, operator=Operator.greater, rhs=split_pt)
        ).and_(AtomicExpression(colname=colname, operator=Operator.not_equal, rhs=None)).current
        scoring.add_split_condition(split_condition, split_point=split_pt)

    if col_stats.has_missing_values:
        condition = AtomicExpression(colname=colname, operator=Operator.not_equal, rhs=None)
        scoring.add_split_condition(condition, np.nan)

    return scoring.get_report()


def find_split_cat(
    data: pl.LazyFrame | pl.DataFrame,
    colname: str,
    y_true: pl.Series,
    min_count: int = 10,
    scoring: BaseSplitScoring | None = None
) -> SplitReport:
    """
    Method for finding a split point, that maximizes variance reduction after the split.

    Parameters
    ----------
    min_count : int, default=10
        Minimum required number of samples with a certain category.
        All categories below this requirement are discarded.

    Returns
    -------
    SplitReport
        Results of splitting.
    """

    scoring = scoring or get_scoring_method(data=data, y_true=y_true, colname=colname)

    col = get_col(data, colname)
    val_counts = col.value_counts(sort=True)
    if len(val_counts) == 1:    # either all missing or constant
        return scoring.get_report()

    for category, count in val_counts.rows():
        if count < min_count:
            continue

        # evaluate split point
        scoring.add_split_condition(
            AtomicExpression(colname=colname, operator=Operator.equal, rhs=category),
            split_point=category,
        )

    return scoring.get_report()



from __future__ import annotations

from dataclasses import dataclass
import polars as pl


class Operator:
    less = "<"
    less_or_equal = "<="
    greater = ">"
    greater_or_equal = ">="
    equal = "=="
    not_equal = "!="


class BaseExpression:
    """
    Base class for all logical expressions
    """

    @property
    def column(self) -> str:
        return self.to_polars().meta.output_name()

    def to_polars(self) -> pl.Expr:
        pass

    def apply(self, data: pl.DataFrame | pl.LazyFrame) -> pl.Series:
        tmp_name = "__expr__"
        res = data.with_columns(self.to_polars().alias(tmp_name)).select(tmp_name)
        if isinstance(res, pl.LazyFrame):
            res = res.collect()
        return res.to_series()

    def _serialize(self) -> dict:
        pass

    @staticmethod
    def _deserialize(data: dict) -> BaseExpression:
        pass

    def serialize(self) -> dict:
        return {
            "type": self.__class__.__name__,
            "data": self._serialize()
        }

    @staticmethod
    def deserialize(data: dict | None) -> BaseExpression | None:
        """
        Deserialize BaseExpression instance, which was serialized by `BaseExpression.serialize`.
        """

        if data is None:
            return None

        functions = {  # deserialization functions depending on the object type. See also `BaseExpression.serialize`.
            AtomicExpression.__name__: AtomicExpression._deserialize,
            AndExpression.__name__: AndExpression._deserialize,
            OrExpression.__name__: OrExpression._deserialize,
            NotExpression.__name__: NotExpression._deserialize,
        }

        if data["type"] in functions:
            return functions[data["type"]](data["data"])

        raise NotImplementedError(data["type"])


@dataclass
class AtomicExpression(BaseExpression):
    """
    Leaf class: Atomic Boolean expression (e.g., A > 5)
    """

    colname: str
    operator: Operator
    rhs: int | float | str | None

    def to_polars(self) -> pl.Expr:
        """
        Returns Polars expression for condition. The expression result is Boolean.

        Args
        ----
        propagate_null : bool, default=True
            Defines how empty values are handled. Polars propagates null and NaNs during comparison.
            This can result is having undefined outcomes when combining conditions.
            In order to avoid nulls in the comparison results we can force boolean outcome when data contains
            missing values.
        """

        if self.operator == Operator.less:
            return pl.col(self.colname).lt(self.rhs)

        if self.operator == Operator.less_or_equal:
            return pl.col(self.colname).le(self.rhs)

        if self.operator == Operator.greater:
            return pl.col(self.colname).gt(self.rhs)

        if self.operator == Operator.greater_or_equal:
            return pl.col(self.colname).ge(self.rhs)

        if self.operator == Operator.equal:
            if self.rhs is not None:
                return pl.col(self.colname).eq(self.rhs)
            else:
                return pl.col(self.colname).is_null()

        if self.operator == Operator.not_equal:
            if self.rhs is not None:
                return pl.col(self.colname).ne(self.rhs)
            else:
                return pl.col(self.colname).is_not_null()

    def not_(self) -> AtomicExpression:
        not_operator = {
            Operator.less: Operator.greater_or_equal,
            Operator.less_or_equal: Operator.greater,
            Operator.greater: Operator.less_or_equal,
            Operator.greater_or_equal: Operator.less,
            Operator.equal: Operator.not_equal,
            Operator.not_equal: Operator.equal,
        }
        return AtomicExpression(colname=self.colname, operator=not_operator[self.operator], rhs=self.rhs)

    def __repr__(self) -> str:
        return f"{self.colname} {self.operator} {self.rhs}"

    def _serialize(self) -> dict:
        return {
            "colname": self.colname,
            "operator": self.operator,
            "rhs": self.rhs,
        }

    @staticmethod
    def _deserialize(data: dict) -> AtomicExpression:
        return AtomicExpression(**data)


# Composite classes for logical connectives
class AndExpression(BaseExpression):
    def __init__(
        self,
        left: BaseExpression,
        right: BaseExpression,
    ):
        self.left = left
        self.right = right

    def to_polars(self) -> pl.Expr:
        return self.left.to_polars() & self.right.to_polars()

    def __repr__(self) -> str:
        return f"({self.left}) AND ({self.right})"

    def _serialize(self) -> dict:
        return {
            "left": self.left.serialize() if self.left is not None else None,
            "right": self.right.serialize() if self.right is not None else None,
        }

    @staticmethod
    def _deserialize(data: dict) -> AndExpression:
        return AndExpression(
            left=BaseExpression.deserialize(data["left"]),
            right=BaseExpression.deserialize(data["right"]),
        )


class OrExpression(BaseExpression):
    def __init__(
        self,
        left: BaseExpression,
        right: BaseExpression,
    ):
        self.left = left
        self.right = right

    def to_polars(self) -> pl.Expr:
        return self.left.to_polars() | self.right.to_polars()

    def __repr__(self) -> str:
        return f"({self.left}) OR ({self.right})"

    def _serialize(self) -> dict:
        return {
            "left": self.left.serialize() if self.left is not None else None,
            "right": self.right.serialize() if self.right is not None else None,
        }

    @staticmethod
    def _deserialize(data: dict) -> OrExpression:
        return OrExpression(
            left=BaseExpression.deserialize(data["left"]),
            right=BaseExpression.deserialize(data["right"]),
        )


class NotExpression(BaseExpression):
    def __init__(
        self,
        expr: BaseExpression,
    ):
        self.expr = expr

    def not_(self) -> BaseExpression:
        return self.expr

    def to_polars(self) -> pl.Expr:
        return ~self.expr.to_polars()

    def __repr__(self) -> str:
        return f"NOT ({self.expr})"

    def _serialize(self) -> dict:
        return {
            "expr": self.expr.serialize() if self.expr is not None else None
        }

    @staticmethod
    def _deserialize(data: dict) -> NotExpression:
        return NotExpression(
            expr=BaseExpression.deserialize(data["expr"])
        )


class ExpressionBuilder(BaseExpression):
    """
    Builder pattern for ... building logical expressions and conditions.
    """

    def __init__(self, start: BaseExpression | None = None):
        self.current = start

    def and_(self, other: BaseExpression) -> ExpressionBuilder:
        if self.current is None:
            raise RuntimeError()

        self.current = AndExpression(self.current, other)
        return self

    def or_(self, other: BaseExpression) -> ExpressionBuilder:
        if self.current is None:
            raise RuntimeError()

        self.current = OrExpression(self.current, other)
        return self

    def not_(self) -> ExpressionBuilder:
        if self.current is None:
            raise RuntimeError()

        if hasattr(self.current, "not_"):
            self.current = self.current.not_()
        else:
            # Proper negation of AND and OR is not implemented yet.
            #   Also, it can make resulting leaf conditions look more confusing
            self.current = NotExpression(self.current)
        return self

    def to_polars(self) -> pl.Expr:
        if self.current is None:
            raise RuntimeError()

        return self.current.to_polars()

    def apply(self, data: pl.DataFrame | pl.LazyFrame) -> pl.Series:
        if self.current is None:
            raise RuntimeError()

        return self.current.apply(data)

    def __repr__(self) -> str:
        if self.current is None:
            return ""

        return str(self.current)



from __future__ import annotations

from dataclasses import dataclass, field
import numpy as np
import polars as pl

from .logic import BaseExpression, ExpressionBuilder, AtomicExpression


@dataclass
class TargetStats:
    score_reduction: float | None = None  # reduction of a score *after* the split at the node.
    mean: float | None = None  # [regression] average value of the target, corresponding to the node.
    var: float | None = None  # [regression] variance of the target, corresponding to the node.
    distr: dict[str, int] | None = None  # [classification] target distribution.


@dataclass
class Node:
    """
    Node for a binary tree.
    """

    depth: int
    condition: BaseExpression | None = None  # the most *recent* conditions that lead to the node.
    right: Node | None = None  # "Yes" branch: corresponds to the `greater` for numerical or `equal` for categorical variables.
    left: Node | None = None  # "No" branch: corresponds to the `less or equal` for numerical or `not equal` for categorical variables.
    parent: Node | None = None
    size: int | None = None  # number of samples from the training data that fall into the node.
    target_stats: TargetStats | None = None

    def __repr__(self) -> str:
        return f"d{self.depth} | {self.size} | {self.target_stats.mean:.6f} | {self.condition}"

    @property
    def full_condition(self) -> BaseExpression:
        """
        Returns all conditions starting from the very top parent node and down to the current node.

        The conditions are sorted by depth.
        """
        if self.parent is None:  # if root node
            return self.condition

        is_left = id(self.parent.left) == id(self)  # checks whether node is located in the left branch, and should negate the full path.
        parent_cond = self.parent.full_condition  # "path" to the current node
        if is_left:  # negate the last condition ("path" segment)
            if isinstance(parent_cond, AtomicExpression):
                parent_cond = ExpressionBuilder(parent_cond).not_().current
            else:  # if not atomic then it is an AND expression
                if self.depth == 2:  # if parent node is a root, negate the full parent condition
                    parent_cond = ExpressionBuilder(parent_cond).not_().current
                else:  # if parent node is not root, then each terms in the AND expression are compound expressions
                    parent_cond.right = ExpressionBuilder(parent_cond.right).not_().current

        if self.condition is not None:  # if non-root non-terminal node
            return ExpressionBuilder(parent_cond).and_(self.condition).current

        return parent_cond

    def serialize(self) -> dict:
        return {
            "depth": self.depth,
            "condition": self.condition.serialize() if self.condition is not None else None,
            "right": self.right.serialize() if self.right is not None else None,
            "left": self.left.serialize() if self.left is not None else None,
            "parent": None,  # always None in order to avoid hitting the recursion limit.
            "size": self.size,
            "target_stats": self.target_stats,
        }

    @staticmethod
    def deserialize(data: dict) -> Node:
        res = Node(
            depth=data["depth"],
            condition=BaseExpression.deserialize(data["condition"]),
            size=data["size"],
            target_stats=TargetStats(**data["target_stats"]) if data["target_stats"] is not None else None,
            left=Node.deserialize(data["left"]) if data["left"] is not None else None,
            right=Node.deserialize(data["right"]) if data["right"] is not None else None,
        )
        if res.left:
            res.left.parent = res
        if res.right:
            res.right.parent = res
        return res


@dataclass
class SplitReport:
    split_pts: list[float | int | str] = field(default_factory=list)
    split_evals: list[float] = field(default_factory=list)
    best_split_condition: BaseExpression | None = None  # split condition, which is True for the *Yes* branch.
    best_idx: int | None = None

    @property
    def colname(self) -> str | None:
        if not self.best_split_condition:
            return None
        return self.best_split_condition.column

    @property
    def best_split_point(self) -> float | int | str | None:
        if self.best_idx is None:
            return None

        return self.split_pts[self.best_idx]

    @property
    def best_split_eval(self) -> float | None:
        if self.best_idx is None:
            return None

        return self.split_evals[self.best_idx]


class BaseSplitScoring:
    data: pl.LazyFrame | pl.DataFrame
    y_true: pl.Series
    column: str
    split_conditions: list[BaseExpression]
    split_scores: list[float]
    split_points: list[int | float | str | None]

    def __init__(self, data: pl.LazyFrame | pl.DataFrame, y_true: pl.Series, column: str):
        self.data = data
        self.y_true = y_true
        self.column = column
        self.split_conditions = []
        self.split_scores = []
        self.split_points = []

    def add_split_condition(self, condition: BaseExpression, split_point: int | float | str | None):
        """
        Add condition for splitting and compute split score.
        """
        pass

    def get_report(self) -> SplitReport:
        best_idx = int(np.argmax(self.split_scores)) if len(self.split_scores) > 0 else None
        return SplitReport(
            split_pts=self.split_points,
            split_evals=self.split_scores,
            best_idx=best_idx,
            best_split_condition=self.split_conditions[best_idx] if best_idx is not None else None,
        )


class VarianceScoring(BaseSplitScoring):
    def __init__(self, data: pl.LazyFrame | pl.DataFrame, y_true: pl.Series, column: str):
        super().__init__(data, y_true, column=column)
        self.variance = self.y_true.var()

    def add_split_condition(self, condition: BaseExpression, split_point: int | float | str | None):
        true_mask = condition.apply(self.data)
        if true_mask.all() or true_mask.sum() <= 1:
            return  # one of the splits is empty or contains only 1 element.
        false_mask = ~true_mask
        if false_mask.all() or false_mask.sum() <= 1:
            return  # one of the splits is empty or contains only 1 element.

        gt_true = self.y_true.filter(true_mask)
        gt_false = self.y_true.filter(false_mask)
        num_false = len(gt_false)
        num_true = len(self.y_true) - num_false

        false_var = gt_false.var() or 0
        true_var = gt_true.var() or 0
        split_var = (false_var * num_false + true_var * num_true) / len(self.y_true)
        reduction = self.variance - split_var
        self.split_conditions.append(condition)
        self.split_scores.append(reduction)
        self.split_points.append(split_point)


class EntropyScoring(BaseSplitScoring):
    @staticmethod
    def _entropy(s: pl.Series) -> float:
        return s.value_counts().select("count").to_series().entropy()

    def __init__(self, data: pl.LazyFrame | pl.DataFrame, y_true: pl.Series, column: str):
        super().__init__(data, y_true, column=column)
        self.entropy = self._entropy(self.y_true)

    def add_split_condition(self, condition: BaseExpression, split_point: int | float | str | None):
        """
        Add condition for splitting and compute split score.
        """
        true_mask = condition.apply(self.data)
        if true_mask.all() or true_mask.sum() <= 1:
            return  # one of the splits is empty or contains only 1 element.
        false_mask = ~true_mask
        if false_mask.all() or false_mask.sum() <= 1:
            return  # one of the splits is empty or contains only 1 element.

        ent_true = self._entropy(self.y_true.filter(true_mask))
        ent_false = self._entropy(self.y_true.filter(false_mask))
        num_false = false_mask.sum()
        num_true = len(self.y_true) - num_false

        reduction = self.entropy - (num_true * ent_true + num_false * ent_false) / len(self.y_true)
        self.split_conditions.append(condition)
        self.split_scores.append(reduction)
        self.split_points.append(split_point)
